/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package androidx.room.integration.testapp.migration;

import static org.hamcrest.CoreMatchers.containsString;
import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.Assert.fail;

import android.database.sqlite.SQLiteException;

import androidx.annotation.NonNull;
import androidx.room.Room;
import androidx.room.migration.Migration;
import androidx.room.testing.MigrationTestHelper;
import androidx.sqlite.db.SupportSQLiteDatabase;
import androidx.test.ext.junit.runners.AndroidJUnit4;
import androidx.test.filters.LargeTest;
import androidx.test.platform.app.InstrumentationRegistry;

import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;

import java.io.IOException;

/**
 * Test custom database migrations.
 */
@RunWith(AndroidJUnit4.class)
@LargeTest
public class AutoMigrationTest {
    private static final String TEST_DB = "auto-migration-test";
    @Rule
    public MigrationTestHelper helper;

    public AutoMigrationTest() {
        helper = new MigrationTestHelper(InstrumentationRegistry.getInstrumentation(),
                AutoMigrationDb.class.getCanonicalName());
    }

    // Run this to create the very 1st version of the db.
    public void createFirstVersion() throws IOException {
        SupportSQLiteDatabase db = helper.createDatabase(TEST_DB, 2);
        db.close();
    }

    /**
     * Tests the case where a non existent auto migration is called.
     */
    @Test
    public void testBadAutoMigrationInput() throws IOException {
        try (SupportSQLiteDatabase db = helper.createDatabase(TEST_DB, 1)) {
            db.execSQL("INSERT INTO Entity1 (id, name) VALUES (1, 'row1')");
            AutoMigrationDb autoMigrationDbV2 = getLatestDb();
            helper.runMigrationsAndValidate(
                    TEST_DB,
                    2,
                    true,
                    autoMigrationDbV2.getAutoGeneratedMigration(3, 4)
            );
            fail();
        } catch (IllegalArgumentException ex) {
            assertThat(
                    ex.getMessage(),
                    is("No AutoMigrations between versions 'from = 3', 'to = "
                    + "4' have been provided. Annotate Database class with @AutoMigration(from = "
                    + "3, to = 4) to generate this AutoMigration.")
            );
        }
    }

    @Test
    public void goFromV1ToV2() throws IOException {
        try (SupportSQLiteDatabase db = helper.createDatabase(TEST_DB, 1)) {
            db.execSQL("INSERT INTO Entity1 (id, name) VALUES (1, 'row1')");
        }
        AutoMigrationDb autoMigrationDbV2 = getLatestDb();
        helper.runMigrationsAndValidate(
                TEST_DB,
                2,
                true
        );
        assertThat(autoMigrationDbV2.dao().getAllEntity1s().size(), is(1));
    }

    /**
     * Verifies that the user defined migration is selected over using an autoMigration.
     */
    @Test
    public void goFromV1ToV2WithUserDefinedMigration() throws IOException {
        try (SupportSQLiteDatabase db = helper.createDatabase(TEST_DB, 1)) {
            db.execSQL("INSERT INTO Entity1 (id, name) VALUES (1, 'row1')");
        }

        try {
            AutoMigrationDb autoMigrationDbV2 = Room.databaseBuilder(
                    InstrumentationRegistry.getInstrumentation().getTargetContext(),
                    AutoMigrationDb.class, TEST_DB).addMigrations(MIGRATION_1_2).build();
            autoMigrationDbV2.getOpenHelper().getWritableDatabase(); // trigger open
            helper.closeWhenFinished(autoMigrationDbV2);

            helper.runMigrationsAndValidate(
                    TEST_DB,
                    2,
                    true
            );
        } catch (SQLiteException exception) {
            assertThat(exception.getMessage(), containsString("no such table: Entity0"));
        }
    }

    private AutoMigrationDb getLatestDb() {
        AutoMigrationDb db = Room.databaseBuilder(
                InstrumentationRegistry.getInstrumentation().getTargetContext(),
                AutoMigrationDb.class, TEST_DB).build();
        db.getOpenHelper().getWritableDatabase(); // trigger open
        helper.closeWhenFinished(db);
        return db;
    }

    private static final Migration MIGRATION_1_2 = new Migration(1, 2) {
        @Override
        public void migrate(@NonNull SupportSQLiteDatabase database) {
            database.execSQL("ALTER TABLE `Entity0` ADD COLUMN `addedInV2` INTEGER NOT NULL "
                    + "DEFAULT 2");
        }
    };
}
